# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Displays info from dabble nodes such as FPS, object count, and zone count in a
legend box.
"""

from typing import Any, Dict, List

from peekingduck.pipeline.nodes.draw.utils.legend import Legend
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Draws legend box and information on image.

    The ``draw.legend`` node dynamically pulls the output results of previous
    nodes and uses it to draw the information into a legend box. Currently
    draws fps, object counts and object count in zones.

    Inputs:
        ``all`` (:obj:`Any`): Receives inputs from all preceding outputs to use as
        dynamic input for legend creation.

    Outputs:
        |none|

    Configs:
        all_legend_items (:obj:`List[str]`):
            **default = ["fps", "count", "zone_count"]**. |br|
            A list of all possible items that can be drawn in legend box. The
            information to be drawn is selected by ``include``. This is done so
            we can have the outputs but choose not to draw them on screen.
        position (:obj:`str`): **{"top", "bottom"}, default = "bottom"**. |br|
            Position to draw legend box. "top" draws it at the top-left
            position while "bottom" draws it at bottom-left.
        include (:obj:`List[str]`): **default = ["all_legend_items"]**. |br|
            List of information to draw. Currently, "fps", "count" and/or
            "zone_count" can be drawn. The default value "all_legend_items"
            draws everything dynamically depending on inputs.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.include: List[str]
        self.legend_items: List[str] = []

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Draws legend box with information from nodes.

        Args:
            inputs (dict): Dictionary with all available keys.

        Returns:
            outputs (dict): Dictionary with keys "none".
        """
        if len(self.legend_items) == 0:
            # Check inputs to set legend items to draw
            if self.include[0] == "all_legend_items":
                self.include = self.all_legend_items
            self._include(inputs)
        if len(self.legend_items) != 0:
            Legend().draw(inputs, self.legend_items, self.position)
        else:
            return {}
        # cv2 weighted does not update the referenced image. Need to return and
        # replace.
        return {"img": inputs["img"]}

    def _include(self, inputs: Dict[str, Any]) -> None:
        for item in self.all_legend_items:
            if item in inputs.keys() and item in self.include:
                self.legend_items.append(item)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Face detection model with model types: yolov4 and yolov4tiny
"""

import logging
from typing import List, Dict, Any, Tuple
import numpy as np
from peekingduck.weights_utils import checker, downloader, finder
from peekingduck.pipeline.nodes.model.yolov4_face.yolo_face_files.detector import (
    Detector,
)


class Yolov4:  # pylint: disable=too-few-public-methods
    """Yolo model with model types: v4 and v4tiny"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__()

        self.logger = logging.getLogger(__name__)

        # check threshold values
        if not 0 <= config["yolo_score_threshold"] <= 1:
            raise ValueError("yolo_score_threshold must be in [0, 1]")

        if not 0 <= config["yolo_iou_threshold"] <= 1:
            raise ValueError("yolo_iou_threshold must be in [0, 1]")

        weights_dir, model_dir = finder.find_paths(
            config["root"], config["weights"], config["weights_parent_dir"]
        )

        # check for yolo weights, if none then download into weights folder
        if not checker.has_weights(weights_dir, model_dir):
            self.logger.info("---no weights detected. proceeding to download...---")
            downloader.download_weights(weights_dir, config["weights"]["blob_file"])
            self.logger.info(f"---weights downloaded to {weights_dir}.---")

        self.detect_ids = config["detect_ids"]
        self.detector = Detector(config, model_dir)

    def predict(self, frame: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Predicts face bboxes, labels and scores

        Args:
            frame (np.ndarray): image in numpy array

        Returns:
            bboxes (np.ndarray): numpy array of detected bboxes
            labels (np.ndarray): numpy array of class labels
            scores (np.ndarray): numpy array of confidence scores
        """
        assert isinstance(frame, np.ndarray)

        return self.detector.predict_object_bbox_from_image(frame)

    def get_detect_ids(self) -> List[int]:
        """Getter for selected ids for detection. This function is used in unit
        testing.

        Returns:
            List[int]: list of selected detection ids
        """
        return self.detect_ids



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Slower but more accurate object detection model.
"""

from typing import Any, Dict

from peekingduck.pipeline.nodes.model.efficientdet_d04 import efficientdet_model
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Initialises an EfficientDet model to detect bounding boxes from an image.

    The EfficientDet node is capable of detecting objects from 80 categories.
    The table of categories can be found
    :ref:`here <general-object-detection-ids>`.

    EfficientDet node has five levels of compound coefficient (0 - 5). A higher
    compound coefficient will scale up all dimensions of the backbone network
    width, depth, and input resolution, which results in better performance but
    slower inference time. The default compound coefficient is 0 and can be
    changed to other values.

    Inputs:
        |img|

    Outputs:
        |bboxes|

        |bbox_labels|

        |bbox_scores|

    Configs:
        model_type (:obj:`int`): **{0, 1, 2, 3, 4}, default = 0**. |br|
            Defines the compound coefficient for EfficientDet.
        score_threshold (:obj:`float`): **[0, 1], default = 0.3**.
            Threshold to determine if detection should be returned.
        detect_ids (:obj:`List[int]`): **default = [0]**. |br|
            List of object class IDs to be detected.
        weights_parent_dir (:obj:`Optional[str]`): **default = null**. |br|
            Change the parent directory where weights will be stored by replacing
            ``null`` with an absolute path to the desired directory.

    References:
        EfficientDet: Scalable and Efficient Object Detection:
        https://arxiv.org/abs/1911.09070

        Code adapted from https://github.com/xuannianz/EfficientDet.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.model = efficientdet_model.EfficientDetModel(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Takes an image as input and returns bboxes of objects specified
        in config.
        """
        # Currently prototyped to return just the bounding boxes
        # without the scores
        bboxes, labels, scores = self.model.predict(inputs["img"])
        outputs = {"bboxes": bboxes, "bbox_labels": labels, "bbox_scores": scores}
        return outputs



# Copyright 2019 The TensorFlow Authors, Pavel Yakubovskiy, Björn Barz. All Rights Reserved.
#
# Modifications copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# Contains definitions for EfficientNet model.
#
# [1] Mingxing Tan, Quoc V. Le
#   EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks.
#   ICML'19, https://arxiv.org/abs/1905.11946
#
# Code of this model implementation is mostly written by
# Björn Barz ([@Callidior](https://github.com/Callidior))

"""
EfficientNet parameters and constants
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import math
import string
import collections
from typing import Any, Dict, Callable, List, Tuple, Union

from six.moves import xrange
import tensorflow as tf
from tensorflow.python.keras.applications.imagenet_utils import obtain_input_shape
from tensorflow.keras.applications.imagenet_utils import (
    preprocess_input as _preprocess_input,
)

from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files.utils.submodule import (  # pylint: disable=line-too-long
    get_submodules_from_kwargs,
)


BACKEND = None
LAYERS = None
MODELS = None
KERAS_UTILS = None


BASE_WEIGHTS_PATH = (
    "https://github.com/Callidior/keras-applications/releases/download/efficientnet/"
)

# fmt: off
WEIGHTS_HASHES = {
    "efficientnet-b0": (
        "163292582f1c6eaca8e7dc7b51b01c61"
        "5b0dbc0039699b4dcd0b975cc21533dc",
        "c1421ad80a9fc67c2cc4000f666aa507"
        "89ce39eedb4e06d531b0c593890ccff3",
    ),
    "efficientnet-b1": (
        "d0a71ddf51ef7a0ca425bab32b7fa7f1"
        "6043ee598ecee73fc674d9560c8f09b0",
        "75de265d03ac52fa74f2f510455ba64f"
        "9c7c5fd96dc923cd4bfefa3d680c4b68",
    ),
    "efficientnet-b2": (
        "bb5451507a6418a574534aa76a91b106"
        "f6b605f3b5dde0b21055694319853086",
        "433b60584fafba1ea3de07443b74cfd3"
        "2ce004a012020b07ef69e22ba8669333",
    ),
    "efficientnet-b3": (
        "03f1fba367f070bd2545f081cfa7f3e7"
        "6f5e1aa3b6f4db700f00552901e75ab9",
        "c5d42eb6cfae8567b418ad3845cfd63a"
        "a48b87f1bd5df8658a49375a9f3135c7",
    ),
    "efficientnet-b4": (
        "98852de93f74d9833c8640474b2c698d"
        "b45ec60690c75b3bacb1845e907bf94f",
        "7942c1407ff1feb34113995864970cd4"
        "d9d91ea64877e8d9c38b6c1e0767c411",
    ),
    "efficientnet-b5": (
        "30172f1d45f9b8a41352d4219bf930ee"
        "3339025fd26ab314a817ba8918fefc7d",
        "9d197bc2bfe29165c10a2af8c2ebc675"
        "07f5d70456f09e584c71b822941b1952",
    ),
    "efficientnet-b6": (
        "f5270466747753485a082092ac9939ca"
        "a546eb3f09edca6d6fff842cad938720",
        "1d0923bb038f2f8060faaf0a0449db4b"
        "96549a881747b7c7678724ac79f427ed",
    ),
    "efficientnet-b7": (
        "876a41319980638fa597acbbf956a82d"
        "10819531ff2dcb1a52277f10c7aefa1a",
        "60b56ff3a8daccc8d96edfd40b204c11"
        "3e51748da657afd58034d54d3cec2bac",
    ),
}
# fmt: on

BlockArgs = collections.namedtuple(
    "BlockArgs",
    [
        "kernel_size",
        "num_repeat",
        "input_filters",
        "output_filters",
        "expand_ratio",
        "id_skip",
        "strides",
        "se_ratio",
    ],
)
# defaults will be a public argument for namedtuple in Python 3.7
# https://docs.python.org/3/library/collections.html#collections.namedtuple
# mypy: ignore-errors
BlockArgs.__new__.__defaults__ = (None,) * len(BlockArgs._fields)  # type: ignore

DEFAULT_BLOCKS_ARGS = (
    BlockArgs(
        kernel_size=3,
        num_repeat=1,
        input_filters=32,
        output_filters=16,
        expand_ratio=1,
        id_skip=True,
        strides=[1, 1],
        se_ratio=0.25,
    ),
    BlockArgs(
        kernel_size=3,
        num_repeat=2,
        input_filters=16,
        output_filters=24,
        expand_ratio=6,
        id_skip=True,
        strides=[2, 2],
        se_ratio=0.25,
    ),
    BlockArgs(
        kernel_size=5,
        num_repeat=2,
        input_filters=24,
        output_filters=40,
        expand_ratio=6,
        id_skip=True,
        strides=[2, 2],
        se_ratio=0.25,
    ),
    BlockArgs(
        kernel_size=3,
        num_repeat=3,
        input_filters=40,
        output_filters=80,
        expand_ratio=6,
        id_skip=True,
        strides=[2, 2],
        se_ratio=0.25,
    ),
    BlockArgs(
        kernel_size=5,
        num_repeat=3,
        input_filters=80,
        output_filters=112,
        expand_ratio=6,
        id_skip=True,
        strides=[1, 1],
        se_ratio=0.25,
    ),
    BlockArgs(
        kernel_size=5,
        num_repeat=4,
        input_filters=112,
        output_filters=192,
        expand_ratio=6,
        id_skip=True,
        strides=[2, 2],
        se_ratio=0.25,
    ),
    BlockArgs(
        kernel_size=3,
        num_repeat=1,
        input_filters=192,
        output_filters=320,
        expand_ratio=6,
        id_skip=True,
        strides=[1, 1],
        se_ratio=0.25,
    ),
)

CONV_KERNEL_INITIALIZER = {
    "class_name": "VarianceScaling",
    "config": {
        "scale": 2.0,
        "mode": "fan_out",
        # EfficientNet actually uses an untruncated normal distribution for
        # initializing conv layers, but keras.initializers.VarianceScaling use
        # a truncated distribution.
        # We decided against a custom initializer for better serializability.
        "distribution": "normal",
    },
}

DENSE_KERNEL_INITIALIZER = {
    "class_name": "VarianceScaling",
    "config": {"scale": 1.0 / 3.0, "mode": "fan_out", "distribution": "uniform"},
}


def preprocess_input(model_input: tf.Tensor, **kwargs: Dict[str, Any]) -> tf.Tensor:
    """Preprocesses model input"""
    kwargs = {
        k: v for k, v in kwargs.items() if k in ["backend", "layers", "models", "utils"]
    }
    return _preprocess_input(model_input, mode="torch", **kwargs)


def get_swish(**kwargs: Dict[str, Any]) -> Callable:
    """Swish activation function: x * sigmoid(x).
    Reference: [Searching for Activation Functions](https://arxiv.org/abs/1710.05941)
    """
    backend, _, _, _ = get_submodules_from_kwargs(kwargs)

    def swish(swish_x: tf.Tensor) -> tf.Tensor:

        if backend.backend() == "tensorflow":
            try:
                # The native TF implementation has a more
                # memory-efficient gradient implementation
                return backend.tf.nn.swish(swish_x)
            except AttributeError:
                pass

        return swish_x * backend.sigmoid(swish_x)

    return swish


def get_dropout(**kwargs: Union[None, Dict[str, Any]]) -> Any:
    """Wrapper over custom dropout. Fix problem of ``None`` shape for tf.keras.
    It is not possible to define FixedDropout class as global object,
    because we do not have modules for inheritance at first time.

    Issue:
        https://github.com/tensorflow/tensorflow/issues/30946
    """
    backend, layers, _, _ = get_submodules_from_kwargs(kwargs)

    class FixedDropout(layers.Dropout):  # pylint: disable=too-few-public-methods
        """Fixed Dropout Class"""

        def _get_noise_shape(self, inputs: tf.Tensor) -> Union[None, Tuple[Any, ...]]:
            if self.noise_shape is None:
                return self.noise_shape

            symbolic_shape = backend.shape(inputs)
            noise_shape = [
                symbolic_shape[axis] if shape is None else shape
                for axis, shape in enumerate(self.noise_shape)
            ]
            return tuple(noise_shape)

    return FixedDropout


def round_filters(filters: float, width_coefficient: float, depth_divisor: int) -> int:
    """Round number of filters based on width multiplier."""

    filters *= width_coefficient
    new_filters = int(filters + depth_divisor / 2) // depth_divisor * depth_divisor
    new_filters = max(depth_divisor, new_filters)
    # Make sure that round down does not go down by more than 10%.
    if new_filters < 0.9 * filters:
        new_filters += depth_divisor
    return int(new_filters)


def round_repeats(repeats: int, depth_coefficient: float) -> int:
    """Round number of repeats based on depth multiplier."""

    return int(math.ceil(depth_coefficient * repeats))


def mb_conv_block(
    inputs: tf.Tensor,  # type:ignore
    block_args,
    activation: Callable,
    drop_rate: float = None,
    prefix: str = "",
) -> tf.Tensor:
    """Mobile Inverted Residual Bottleneck."""

    has_se = (block_args.se_ratio is not None) and (0 < block_args.se_ratio <= 1)
    bn_axis = 3 if BACKEND.image_data_format() == "channels_last" else 1

    # workaround over non working dropout with None in noise_shape in tf.keras
    Dropout = get_dropout(
        backend=BACKEND, layers=LAYERS, models=MODELS, utils=KERAS_UTILS
    )

    # Expansion phase
    filters = block_args.input_filters * block_args.expand_ratio
    if block_args.expand_ratio != 1:
        x_in = LAYERS.Conv2D(
            filters,
            1,
            padding="same",
            use_bias=False,
            kernel_initializer=CONV_KERNEL_INITIALIZER,
            name=prefix + "expand_conv",
        )(inputs)
        x_in = LAYERS.BatchNormalization(axis=bn_axis, name=prefix + "expand_bn")(x_in)
        x_in = LAYERS.Activation(activation, name=prefix + "expand_activation")(x_in)
    else:
        x_in = inputs

    # Depthwise Convolution
    x_in = LAYERS.DepthwiseConv2D(
        block_args.kernel_size,
        strides=block_args.strides,
        padding="same",
        use_bias=False,
        depthwise_initializer=CONV_KERNEL_INITIALIZER,
        name=prefix + "dwconv",
    )(x_in)
    x_in = LAYERS.BatchNormalization(axis=bn_axis, name=prefix + "bn")(x_in)
    x_in = LAYERS.Activation(activation, name=prefix + "activation")(x_in)

    # Squeeze and Excitation phase
    if has_se:
        num_reduced_filters = max(
            1, int(block_args.input_filters * block_args.se_ratio)
        )
        se_tensor = LAYERS.GlobalAveragePooling2D(name=prefix + "se_squeeze")(x_in)

        target_shape = (
            (1, 1, filters)
            if BACKEND.image_data_format() == "channels_last"
            else (filters, 1, 1)
        )
        se_tensor = LAYERS.Reshape(target_shape, name=prefix + "se_reshape")(se_tensor)
        se_tensor = LAYERS.Conv2D(
            num_reduced_filters,
            1,
            activation=activation,
            padding="same",
            use_bias=True,
            kernel_initializer=CONV_KERNEL_INITIALIZER,
            name=prefix + "se_reduce",
        )(se_tensor)
        se_tensor = LAYERS.Conv2D(
            filters,
            1,
            activation="sigmoid",
            padding="same",
            use_bias=True,
            kernel_initializer=CONV_KERNEL_INITIALIZER,
            name=prefix + "se_expand",
        )(se_tensor)
        if BACKEND.backend() == "theano":
            # For the Theano backend, we have to explicitly make
            # the excitation weights broadcastable.
            pattern = (
                [True, True, True, False]
                if BACKEND.image_data_format() == "channels_last"
                else [True, False, True, True]
            )
            se_tensor = LAYERS.Lambda(
                lambda x: BACKEND.pattern_broadcast(x, pattern),
                name=prefix + "se_broadcast",
            )(se_tensor)
        x_in = LAYERS.multiply([x_in, se_tensor], name=prefix + "se_excite")

    # Output phase
    x_in = LAYERS.Conv2D(
        block_args.output_filters,
        1,
        padding="same",
        use_bias=False,
        kernel_initializer=CONV_KERNEL_INITIALIZER,
        name=prefix + "project_conv",
    )(x_in)
    x_in = LAYERS.BatchNormalization(axis=bn_axis, name=prefix + "project_bn")(x_in)
    if (
        block_args.id_skip
        and all(s == 1 for s in block_args.strides)
        and block_args.input_filters == block_args.output_filters
    ):
        if drop_rate and (drop_rate > 0):
            x_in = Dropout(
                drop_rate, noise_shape=(None, 1, 1, 1), name=prefix + "drop"
            )(x_in)
        x_in = LAYERS.add([x_in, inputs], name=prefix + "add")

    return x_in


def efficientnet_base(  # pylint: disable=too-many-arguments, too-many-locals, too-many-branches
    width_coefficient: float,
    depth_coefficient: float,
    default_resolution: int,
    drop_connect_rate: float = 0.2,
    depth_divisor: int = 8,
    blocks_args=DEFAULT_BLOCKS_ARGS,
    include_top: bool = True,
    weights: str = "imagenet",
    input_tensor: Union[None, tf.Tensor] = None,
    input_shape: Union[None, Tuple[int, int, int]] = None,
    classes: int = 1000,
    **kwargs: Any
) -> List[tf.Tensor]:
    """Instantiates the EfficientNet architecture using given scaling coefficients.
    Optionally loads weights pre-trained on ImageNet.
    Note that the data format convention used by the model is
    the one specified in your Keras config at `~/.keras/keras.json`.
    # Arguments
        width_coefficient: float, scaling coefficient for network width.
        depth_coefficient: float, scaling coefficient for network depth.
        default_resolution: int, default input image size.
        dropout_rate: float, dropout rate before final classifier layer.
        drop_connect_rate: float, dropout rate at skip connections.
        depth_divisor: int.
        blocks_args: A tuple of BlockArgs to construct block modules.
        model_name: string, model name.
        include_top: whether to include the fully-connected
            layer at the top of the network.
        weights: one of `None` (random initialization),
              'imagenet' (pre-training on ImageNet),
              or the path to the weights file to be loaded.
        input_tensor: optional Keras tensor
            (i.e. output of `layers.Input()`)
            to use as image input for the model.
        input_shape: optional shape tuple, only to be specified
            if `include_top` is False.
            It should have exactly 3 inputs channels.
        pooling: optional pooling mode for feature extraction
            when `include_top` is `False`.
            - `None` means that the output of the model will be
                the 4D tensor output of the
                last convolutional layer.
            - `avg` means that global average pooling
                will be applied to the output of the
                last convolutional layer, and thus
                the output of the model will be a 2D tensor.
            - `max` means that global max pooling will
                be applied.
        classes: optional number of classes to classify images
            into, only to be specified if `include_top` is True, and
            if no `weights` argument is specified.
    # Returns
        A Keras model instance.
    # Raises
        ValueError: in case of invalid argument for `weights`,
            or invalid input shape.
    """
    global BACKEND, LAYERS, MODELS, KERAS_UTILS  # pylint: disable=global-statement
    BACKEND, LAYERS, MODELS, KERAS_UTILS = get_submodules_from_kwargs(kwargs)
    features = []
    if not (weights in {"imagenet", None} or os.path.exists(weights)):
        raise ValueError(
            "The `weights` argument should be either "
            "`None` (random initialization), `imagenet` "
            "(pre-training on ImageNet), "
            "or the path to the weights file to be loaded."
        )

    if weights == "imagenet" and include_top and classes != 1000:
        raise ValueError(
            'If using `weights` as `"imagenet"` with `include_top`'
            " as true, `classes` should be 1000"
        )

    # Determine proper input shape
    input_shape = obtain_input_shape(
        input_shape,
        default_size=default_resolution,
        min_size=32,
        data_format=BACKEND.image_data_format(),
        require_flatten=include_top,
        weights=weights,
    )

    if input_tensor is None:
        img_input = LAYERS.Input(shape=input_shape)
    else:
        if BACKEND.backend() == "tensorflow":
            from tensorflow.python.keras.backend import (  # pylint: disable=import-outside-toplevel
                is_keras_tensor,
            )
        else:
            is_keras_tensor = BACKEND.is_keras_tensor
        if not is_keras_tensor(input_tensor):
            img_input = LAYERS.Input(tensor=input_tensor, shape=input_shape)
        else:
            img_input = input_tensor

    bn_axis = 3 if BACKEND.image_data_format() == "channels_last" else 1
    activation = get_swish(**kwargs)

    # Build stem
    x_in = img_input
    x_in = LAYERS.Conv2D(
        round_filters(32, width_coefficient, depth_divisor),
        3,
        strides=(2, 2),
        padding="same",
        use_bias=False,
        kernel_initializer=CONV_KERNEL_INITIALIZER,
        name="stem_conv",
    )(x_in)
    x_in = LAYERS.BatchNormalization(axis=bn_axis, name="stem_bn")(x_in)
    x_in = LAYERS.Activation(activation, name="stem_activation")(x_in)
    # Build blocks
    num_blocks_total = sum(block_args.num_repeat for block_args in blocks_args)
    block_num = 0
    for idx, block_args in enumerate(blocks_args):
        assert block_args.num_repeat > 0
        # Update block input and output filters based on depth multiplier.
        block_args = block_args._replace(
            input_filters=round_filters(
                block_args.input_filters, width_coefficient, depth_divisor
            ),
            output_filters=round_filters(
                block_args.output_filters, width_coefficient, depth_divisor
            ),
            num_repeat=round_repeats(block_args.num_repeat, depth_coefficient),
        )

        # The first block needs to take care of stride and filter size increase.
        drop_rate = drop_connect_rate * float(block_num) / num_blocks_total
        x_in = mb_conv_block(
            x_in,
            block_args,
            activation=activation,
            drop_rate=drop_rate,
            prefix="block{}a_".format(idx + 1),
        )
        block_num += 1
        if block_args.num_repeat > 1:
            # pylint: disable=protected-access
            block_args = block_args._replace(
                input_filters=block_args.output_filters, strides=[1, 1]
            )
            # pylint: enable=protected-access
            for bidx in xrange(block_args.num_repeat - 1):
                drop_rate = drop_connect_rate * float(block_num) / num_blocks_total
                block_prefix = "block{}{}_".format(
                    idx + 1, string.ascii_lowercase[bidx + 1]
                )
                x_in = mb_conv_block(
                    x_in,
                    block_args,
                    activation=activation,
                    drop_rate=drop_rate,
                    prefix=block_prefix,
                )
                block_num += 1
        if idx < len(blocks_args) - 1 and blocks_args[idx + 1].strides[0] == 2:
            features.append(x_in)
        elif idx == len(blocks_args) - 1:
            features.append(x_in)
    return features


def efficientnet_b0(
    include_top: bool = True,
    weights: str = "imagenet",
    input_tensor: Union[None, tf.Tensor] = None,
    input_shape: Union[None, Tuple[int, int, int]] = None,
    classes: int = 1000,
    **kwargs: Any
) -> List[tf.Tensor]:
    """EfficientNet-B0 model"""
    return efficientnet_base(
        1.0,
        1.0,
        224,
        0.2,
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        classes=classes,
        **kwargs
    )


def efficientnet_b1(
    include_top: bool = True,
    weights: str = "imagenet",
    input_tensor: Union[None, tf.Tensor] = None,
    input_shape: Union[None, Tuple[int, int, int]] = None,
    classes: int = 1000,
    **kwargs: Any
) -> List[tf.Tensor]:
    """EfficientNet-B1 model"""
    return efficientnet_base(
        1.0,
        1.1,
        240,
        0.2,
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        classes=classes,
        **kwargs
    )


def efficientnet_b2(
    include_top: bool = True,
    weights: str = "imagenet",
    input_tensor: Union[None, tf.Tensor] = None,
    input_shape: Union[None, Tuple[int, int, int]] = None,
    classes: int = 1000,
    **kwargs: Any
) -> List[tf.Tensor]:
    """EfficientNet-B2 model"""
    return efficientnet_base(
        1.1,
        1.2,
        260,
        0.3,
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        classes=classes,
        **kwargs
    )


def efficientnet_b3(
    include_top: bool = True,
    weights: str = "imagenet",
    input_tensor: Union[None, tf.Tensor] = None,
    input_shape: Union[None, Tuple[int, int, int]] = None,
    classes: int = 1000,
    **kwargs: Any
) -> List[tf.Tensor]:
    """EfficientNet-B3 model"""
    return efficientnet_base(
        1.2,
        1.4,
        300,
        0.3,
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        classes=classes,
        **kwargs
    )


def efficientnet_b4(
    include_top: bool = True,
    weights: str = "imagenet",
    input_tensor: Union[None, tf.Tensor] = None,
    input_shape: Union[None, Tuple[int, int, int]] = None,
    classes: int = 1000,
    **kwargs: Any
) -> List[tf.Tensor]:
    """EfficientNet-B4 model"""
    return efficientnet_base(
        1.4,
        1.8,
        380,
        0.4,
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        classes=classes,
        **kwargs
    )


def efficientnet_b5(
    include_top: bool = True,
    weights: str = "imagenet",
    input_tensor: Union[None, tf.Tensor] = None,
    input_shape: Union[None, Tuple[int, int, int]] = None,
    classes: int = 1000,
    **kwargs: Any
) -> List[tf.Tensor]:
    """EfficientNet-B5 model"""
    return efficientnet_base(
        1.6,
        2.2,
        456,
        0.4,
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        classes=classes,
        **kwargs
    )


def efficientnet_b6(
    include_top: bool = True,
    weights: str = "imagenet",
    input_tensor: Union[None, tf.Tensor] = None,
    input_shape: Union[None, Tuple[int, int, int]] = None,
    classes: int = 1000,
    **kwargs: Any
) -> List[tf.Tensor]:
    """EfficientNet-B6 model"""
    return efficientnet_base(
        1.8,
        2.6,
        528,
        0.5,
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        classes=classes,
        **kwargs
    )


def efficientnet_b7(
    include_top: bool = True,
    weights: str = "imagenet",
    input_tensor: Union[None, tf.Tensor] = None,
    input_shape: Union[None, Tuple[int, int, int]] = None,
    classes: int = 1000,
    **kwargs: Any
) -> List[tf.Tensor]:
    """EfficientNet-B7 model"""
    return efficientnet_base(
        2.0,
        3.1,
        600,
        0.5,
        include_top=include_top,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        classes=classes,
        **kwargs
    )


setattr(efficientnet_b0, "__doc__", efficientnet_base.__doc__)
setattr(efficientnet_b1, "__doc__", efficientnet_base.__doc__)
setattr(efficientnet_b2, "__doc__", efficientnet_base.__doc__)
setattr(efficientnet_b3, "__doc__", efficientnet_base.__doc__)
setattr(efficientnet_b4, "__doc__", efficientnet_base.__doc__)
setattr(efficientnet_b5, "__doc__", efficientnet_base.__doc__)
setattr(efficientnet_b6, "__doc__", efficientnet_base.__doc__)
setattr(efficientnet_b7, "__doc__", efficientnet_base.__doc__)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Reads a videofeed from a stream, e.g., webcam.
"""

from typing import Any, Dict, Union

from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.input.utils.preprocess import resize_image
from peekingduck.pipeline.nodes.input.utils.read import VideoNoThread, VideoThread


class Node(AbstractNode):
    """Receives livestream as inputs.

    Inputs:
        |none|

    Outputs:
        |img|

        |filename|

        |pipeline_end|

        |saved_video_fps|

    Configs:
        fps_saved_output_video (:obj:`int`): **default = 10**. |br|
            FPS of the MP4 file after livestream is processed and exported. FPS
            is dependent on running machine performance.
        filename (:obj:`str`): **default = "webcamfeed.mp4"**. |br|
            Filename of the MP4 file if livestream is exported.
        resize (:obj:`Dict`):
            **default = { do_resizing: False, width: 1280, height: 720 }** |br|
            Dimension of extracted image frame.
        input_source (:obj:`int`): **default = 0**  (for webcam). |br|
            Refer to `OpenCV documentation <https://docs.opencv.org/3.4/d4/d15/
            group__videoio__flags__base.html#ga023786be1ee68a9105bf2e48c700294d
            />`_ for list of source stream codes.
        mirror_image (:obj:`bool`): **default = False**. |br|
            Flag to set extracted image frame as mirror image of input stream.
        frames_log_freq (:obj:`int`): **default = 100**. |br|
            Logs frequency of frames passed in CLI
        threading (:obj:`bool`): **default = False**. |br|
            Flag to enable threading when reading frames from camera.
            The FPS can increase up to 30%.
        buffer_frames (:obj:`bool`): **default = False**. |br|
            Boolean to indicate if threaded class should buffer image frames.
            If threading is True and output.media_writer is enabled, then
            buffer_frames should be True to ensure output video is correctly
            saved. One side effect of threading=True, buffer_frames=True is the
            onscreen video display could appear laggy due to the buffering. |br|
            For more info, please refer to `input.live configuration
            <https://github.com/aimakerspace/PeekingDuck/blob/dev/peekingduck/configs/input/live.yml>`_.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self._allowed_extensions = ["mp4", "avi", "mov", "mkv"]
        self.videocap: Union[VideoNoThread, VideoThread]
        if self.threading:
            self.videocap = VideoThread(
                self.input_source, self.mirror_image, self.buffer_frames
            )
        else:
            self.videocap = VideoNoThread(self.input_source, self.mirror_image)

        width, height = self.videocap.resolution
        self.logger.info(f"Device resolution used: {width} by {height}")
        if self.resize["do_resizing"]:
            self.logger.info(
                f"Resizing of input set to {self.resize['width']} "
                f"by {self.resize['height']}"
            )
        if self.filename.split(".")[-1] not in self._allowed_extensions:
            raise ValueError(
                f"filename extension must be one of: {self._allowed_extensions}"
            )

        self.frame_counter = 0

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        success, img = self.videocap.read_frame()

        if success:
            if self.resize["do_resizing"]:
                img = resize_image(img, self.resize["width"], self.resize["height"])

            outputs = {
                "img": img,
                "pipeline_end": False,
                "filename": self.filename,
                "saved_video_fps": self.fps_saved_output_video,
            }
            self.frame_counter += 1
            if self.frame_counter % self.frames_log_freq == 0:
                self.logger.info(f"Frames Processed: {self.frame_counter} ...")

        else:
            outputs = {
                "img": None,
                "pipeline_end": True,
                "filename": self.filename,
                "saved_video_fps": self.fps_saved_output_video,
            }
            self.logger.warning("No video frames available for processing.")

        return outputs

    def release_resources(self) -> None:
        """Override base class method to free video resource"""
        self.videocap.shutdown()



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
functions for drawing bounding box related UI components
"""

from typing import List, Tuple

import cv2
import numpy as np
from cv2 import FONT_HERSHEY_SIMPLEX, LINE_AA

from peekingduck.pipeline.nodes.draw.utils.constants import (
    BLACK,
    CHAMPAGNE,
    FILLED,
    NORMAL_FONTSCALE,
    POINT_RADIUS,
    PRIMARY_PALETTE,
)
from peekingduck.pipeline.nodes.draw.utils.constants import (
    PRIMARY_PALETTE_LENGTH as TOTAL_COLOURS,
)
from peekingduck.pipeline.nodes.draw.utils.constants import THICK, VERY_THICK
from peekingduck.pipeline.nodes.draw.utils.general import (
    get_image_size,
    project_points_onto_original_image,
)


def draw_bboxes(
    frame: np.ndarray,
    bboxes: List[List[float]],
    bbox_labels: List[str],
    show_labels: bool,
    colour_choice: Tuple[int, int, int] = None,
) -> None:
    """Draw bboxes onto an image frame.

    Args:
        frame (np.array): image of current frame
        bboxes (List[List[float]]): bounding box coordinates
        colour (Tuple[int, int, int]): colour used for bounding box
        bbox_labels (List[str]): labels of object detected
    """
    image_size = get_image_size(frame)
    # Get unique label colour indexes
    colour_indx = {label: indx for indx, label in enumerate(set(bbox_labels))}

    for i, bbox in enumerate(bboxes):
        if colour_choice:
            colour = colour_choice
        else:
            colour = PRIMARY_PALETTE[colour_indx[bbox_labels[i]] % TOTAL_COLOURS]
        if show_labels:
            _draw_bbox(frame, bbox, image_size, colour, bbox_labels[i])
        else:
            _draw_bbox(frame, bbox, image_size, colour)


def _draw_bbox(
    frame: np.ndarray,
    bbox: np.ndarray,
    image_size: Tuple[int, int],
    colour: Tuple[int, int, int],
    bbox_label: str = None,
) -> np.ndarray:
    """Draw a single bounding box"""
    top_left, bottom_right = project_points_onto_original_image(bbox, image_size)
    cv2.rectangle(
        frame,
        (top_left[0], top_left[1]),
        (bottom_right[0], bottom_right[1]),
        colour,
        VERY_THICK,
    )

    if bbox_label:
        _draw_label(frame, top_left, bbox_label, colour, BLACK)


def _draw_label(
    frame: np.ndarray,
    top_left: Tuple[int, int],
    bbox_label: str,
    bg_colour: Tuple[int, int, int],
    text_colour: Tuple[int, int, int],
) -> None:
    """Draw bbox label at top left of bbox"""
    # get label size
    (text_width, text_height), baseline = cv2.getTextSize(
        bbox_label, FONT_HERSHEY_SIMPLEX, NORMAL_FONTSCALE, THICK
    )
    # put filled text rectangle
    cv2.rectangle(
        frame,
        (top_left[0], top_left[1]),
        (top_left[0] + text_width, top_left[1] - text_height - baseline),
        bg_colour,
        FILLED,
    )

    # put text above rectangle
    bbox_label = bbox_label[:1].capitalize() + bbox_label[1:]
    cv2.putText(
        frame,
        bbox_label,
        (top_left[0], top_left[1] - 6),
        FONT_HERSHEY_SIMPLEX,
        NORMAL_FONTSCALE,
        text_colour,
        THICK,
        LINE_AA,
    )


def draw_tags(
    frame: np.ndarray,
    bboxes: np.ndarray,
    tags: List[str],
    colour: Tuple[int, int, int],
) -> None:
    """Draw tags above bboxes.

    Args:
        frame (np.ndarray): image of current frame
        bboxes (np.ndarray): bounding box coordinates
        tags (List[string]): tag associated with bounding box
        color (Tuple[int, int, int]): color of text
    """
    image_size = get_image_size(frame)
    for idx, bbox in enumerate(bboxes):
        _draw_tag(frame, bbox, tags[idx], image_size, colour)


def _draw_tag(
    frame: np.ndarray,
    bbox: np.ndarray,
    tag: str,
    image_size: Tuple[int, int],
    colour: Tuple[int, int, int],
) -> None:
    """Draw a tag above a single bounding box."""
    top_left, btm_right = project_points_onto_original_image(bbox, image_size)

    # Find offset to centralize text
    (text_width, _), baseline = cv2.getTextSize(
        tag, FONT_HERSHEY_SIMPLEX, NORMAL_FONTSCALE, THICK
    )
    bbox_width = btm_right[0] - top_left[0]
    offset = int((bbox_width - text_width) / 2)
    position = (top_left[0] + offset, top_left[1] - baseline)
    cv2.putText(
        frame, tag, position, FONT_HERSHEY_SIMPLEX, NORMAL_FONTSCALE, colour, VERY_THICK
    )


def draw_pts(frame: np.ndarray, pts: List[Tuple[float]]) -> None:
    """draw pts of selected object onto frame

    Args:
        frame (np.array): image of current frame
        pts (List[Tuple[float]]): bottom midpoints of bboxes
    """
    for point in pts:
        cv2.circle(frame, point, POINT_RADIUS, CHAMPAGNE, -1)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Yolo model with model types: v3 and v3tiny
"""

import logging
from typing import List, Dict, Any, Tuple

import numpy as np

from peekingduck.pipeline.nodes.model.yolov4.yolo_files.detector import Detector
from peekingduck.weights_utils import checker, downloader, finder


class YoloModel:
    """Yolo model with model types: v4 and v4tiny"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__()
        self.logger = logging.getLogger(__name__)

        # check threshold values
        if not 0 <= config["yolo_score_threshold"] <= 1:
            raise ValueError("yolo_score_threshold must be in [0, 1]")

        weights_dir, model_dir = finder.find_paths(
            config["root"], config["weights"], config["weights_parent_dir"]
        )

        # check for yolo weights, if none then download into weights folder
        if not checker.has_weights(weights_dir, model_dir):
            self.logger.info("---no weights detected. proceeding to download...---")
            downloader.download_weights(weights_dir, config["weights"]["blob_file"])
            self.logger.info(f"---weights downloaded to {weights_dir}.---")

        classes_path = model_dir / config["weights"]["classes_file"]

        with open(classes_path) as infile:
            self.class_names = [c.strip() for c in infile.readlines()]
        self.detect_ids = config["detect_ids"]

        self.detector = Detector(config, model_dir)

    def predict(
        self, frame: np.ndarray
    ) -> Tuple[List[np.ndarray], List[str], List[float]]:
        """predict the bbox from frame

        Returns:
            object_bboxes (List[Numpy Array]): list of bboxes detected
            object_labels (List[str]): list of string labels of the
                object detected for the corresponding bbox
            object_scores (List(float)): list of confidence scores of the
                object detected for the corresponding bbox
        """
        assert isinstance(frame, np.ndarray)

        # return bboxes, object_bboxes, object_labels, object_scores
        return self.detector.predict_object_bbox_from_image(
            self.class_names, frame, self.detect_ids
        )

    def get_detect_ids(self) -> List[int]:
        """getter for selected ids for detection

        Returns:
            List[int]: list of selected detection ids
        """
        return self.detect_ids



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Assigns objects in close proximity to groups.
"""

from typing import Any, Dict, List, Tuple

import numpy as np

from peekingduck.pipeline.nodes.dabble.utils.quick_find import QuickFind
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Groups objects that are near each other.

    It does so by comparing the 3D location of all objects, and assigning
    objects near each other to the same group.

    Inputs:
        |obj_3D_locs|

    Outputs:
        |obj_groups|

    Configs:
        obj_dist_thres (:obj:`float`): **default = 1.5**. |br|
            Threshold of distance, in metres, between two objects. Objects with
            distance less than ``obj_dist_thres`` would be assigned to the same
            group.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Checks the distance between 3d locations of a pair of objects.

        If distance is less than threshold, assign the objects to the same
        group. Repeat for all object pairs.
        """
        nearby_obj_pairs = self._find_nearby_obj_pairs(
            inputs["obj_3D_locs"], self.obj_dist_thres
        )

        quickfind = QuickFind(len(inputs["obj_3D_locs"]))
        for (idx_1, idx_2) in nearby_obj_pairs:
            if not quickfind.connected(idx_1, idx_2):
                quickfind.union(idx_1, idx_2)

        return {"obj_groups": quickfind.get_group_alloc()}

    @staticmethod
    def _find_nearby_obj_pairs(
        obj_locs: List[np.ndarray], obj_dist_thres: float
    ) -> List[Tuple[int, int]]:
        """If the distance between 2 objects are less than the threshold,
        append their indexes to nearby_obj_pairs as a tuple.
        """
        nearby_obj_pairs = []
        for idx_1, loc_1 in enumerate(obj_locs):
            for idx_2, loc_2 in enumerate(obj_locs):
                if idx_1 == idx_2:
                    continue

                dist_bet = np.linalg.norm(loc_1 - loc_2)

                if dist_bet <= obj_dist_thres:
                    if (idx_2, idx_1) not in nearby_obj_pairs:
                        nearby_obj_pairs.append((idx_1, idx_2))

        return nearby_obj_pairs



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Main class for MTCNN Model
"""

import logging
from typing import Dict, Any, Tuple

import numpy as np

from peekingduck.weights_utils import checker, downloader, finder
from peekingduck.pipeline.nodes.model.mtcnnv1.mtcnn_files.detector import Detector


class MtcnnModel:  # pylint: disable=too-few-public-methods
    """MTCNN model to detect face bboxes and landmarks"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__()

        self.logger = logging.getLogger(__name__)

        # check factor value
        if not 0 <= config["mtcnn_factor"] <= 1:
            raise ValueError("mtcnn_factor must be between 0 and 1")

        # check threshold values
        for threshold in config["mtcnn_thresholds"]:
            if not 0 <= threshold <= 1:
                raise ValueError("mtcnn_thresholds must be between 0 and 1")

        # check score value
        if not 0 <= config["mtcnn_score"] <= 1:
            raise ValueError("mtcnn_score must be between 0 and 1")

        weights_dir, model_dir = finder.find_paths(
            config["root"], config["weights"], config["weights_parent_dir"]
        )

        # check for mtcnn weights, if none then download into weights folder
        if not checker.has_weights(weights_dir, model_dir):
            self.logger.info("---no weights detected. proceeding to download...---")
            downloader.download_weights(weights_dir, config["weights"]["blob_file"])
            self.logger.info(f"---weights downloaded to {weights_dir}.---")

        self.detector = Detector(config, model_dir)

    def predict(
        self, frame: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Predicts face bboxes, scores and landmarks

        Args:
            frame (np.ndarray): image in numpy array

        Returns:
            bboxes (np.ndarray): numpy array of detected bboxes
            scores (np.ndarray): numpy array of confidence scores
            landmarks (np.ndarray): numpy array of facial landmarks
            labels (np.ndarray): numpy array of class labels (i.e. face)
        """
        assert isinstance(frame, np.ndarray)

        # return bboxes, scores, landmarks amd class labels
        return self.detector.predict_bbox_landmarks(frame)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Reads video/images from a directory.
"""

from pathlib import Path
from typing import Any, Dict

from peekingduck.pipeline.nodes.input.utils.preprocess import resize_image
from peekingduck.pipeline.nodes.input.utils.read import (
    VideoThread,
    VideoNoThread,
)
from peekingduck.pipeline.nodes.node import AbstractNode


# pylint: disable=R0902
class Node(AbstractNode):
    """Receives videos/image as inputs.

    Inputs:
        |none|

    Outputs:
        |img|

        |pipeline_end|

        |filename|

        |saved_video_fps|

    Configs:
        resize (:obj:`Dict`):
            **default = { do_resizing: False, width: 1280, height: 720 }**. |br|
            Dimension of extracted image frame.
        input_dir (:obj:`str`): **default = "PeekingDuck/data/input"**. |br|
            The directory to look for recorded video files and images.
        mirror_image (:obj:`bool`): **default = False**. |br|
            Flag to set extracted image frame as mirror image of input stream.
        threading (:obj:`bool`): **default = False**. |br|
            Boolean to enable threading when reading frames from input.
            The FPS may increase if this is enabled (system dependent).
        buffer_frames (:obj:`bool`): **default = False**. |br|
            Boolean to indicate if threaded class should buffer image frames.
            If threading is True, it is highly recommended that buffer_frames is
            also True to avoid losing frames, as otherwise the input thread would
            very likely read ahead of the main thread. |br|
            For more info, please refer to `input.recorded configuration
            <https://github.com/aimakerspace/PeekingDuck/blob/dev/peekingduck/configs/input/recorded.yml>`_.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self._allowed_extensions = [
            "jpg",
            "jpeg",
            "png",
            "mp4",
            "avi",
            "mov",
            "mkv",
        ]
        self.file_end = False
        self.frame_counter = -1
        self.tens_counter = 10
        self._get_files(Path(self.input_dir))
        self._get_next_input()

        width, height = self.videocap.resolution
        self.logger.info(f"Video/Image size: {width} by {height}")
        if self.resize["do_resizing"]:
            self.logger.info(
                f"Resizing of input set to {self.resize['width']} "
                f"by {self.resize['height']}"
            )

        self.logger.info(f"Filepath used: {self.input_dir}")

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """
        input: ["none"],
        output: ["img", "pipeline_end"]
        """
        outputs = self._run_single_file()

        approx_processed = round((self.frame_counter / self.videocap.frame_count) * 100)
        self.frame_counter += 1

        if approx_processed > self.tens_counter:
            self.logger.info(f"Approximately Processed: {self.tens_counter}% ...")
            self.tens_counter += 10

        if self.file_end:
            self.logger.info(f"Completed processing file: {self._file_name}")
            pct_complete = round(100 * self.frame_counter / self.videocap.frame_count)
            self.logger.debug(f"#frames={self.frame_counter}, done={pct_complete}%")
            self._get_next_input()
            outputs = self._run_single_file()
            self.frame_counter = 0
            self.tens_counter = 10

        return outputs

    def _run_single_file(self) -> Dict[str, Any]:
        success, img = self.videocap.read_frame()  # type: ignore

        self.file_end = True
        outputs = {
            "img": None,
            "pipeline_end": True,
            "filename": self._file_name,
            "saved_video_fps": self._fps,
        }
        if success:
            self.file_end = False
            if self.resize["do_resizing"]:
                img = resize_image(img, self.resize["width"], self.resize["height"])
            outputs = {
                "img": img,
                "pipeline_end": False,
                "filename": self._file_name,
                "saved_video_fps": self._fps,
            }

        return outputs

    def _get_files(self, path: Path) -> None:
        self._filepaths = [path]

        if path.is_dir():
            self._filepaths = list(path.iterdir())
            self._filepaths.sort()

        if not path.exists():
            raise FileNotFoundError("Filepath does not exist")
        if not self._filepaths:
            raise FileNotFoundError("No Media files available")

    def _get_next_input(self) -> None:
        if self._filepaths:
            file_path = self._filepaths.pop(0)
            self._file_name = file_path.name

            if self._is_valid_file_type(file_path):
                if getattr(self, "threading", False):
                    self.videocap = VideoThread(  # type: ignore
                        str(file_path), self.mirror_image, self.buffer_frames
                    )
                else:
                    self.videocap = VideoNoThread(  # type: ignore
                        str(file_path), self.mirror_image
                    )
                self._fps = self.videocap.fps
            else:
                self.logger.warning(
                    f"Skipping '{file_path}' as it is not an accepted "
                    f"file format {str(self._allowed_extensions)}"
                )
                self._get_next_input()

    def _is_valid_file_type(self, filepath: Path) -> bool:
        if filepath.suffix[1:] in self._allowed_extensions:
            return True
        return False

    def release_resources(self) -> None:
        """Override base class method to free video resource"""
        self.videocap.shutdown()



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Adjusts brightness of an incoming image.
"""


from typing import Any, Dict

import cv2

from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Changes image contrast and brightness.

    The ``draw.image_processor`` node adjusts contrast and brightness of the
    given image. Uses algorithm by OpenCV. An article providing a good overview
    of the algorithm can be found `here <https://programmer.ink/think/
    adjusting-the-brightness-and-contrast-of-an-image-with-opencv4.3.0-tutorial
    .html#3、API-convertScaleAbs>`_.

    Inputs:
        |img|

    Outputs:
        |img|

    Configs:
        brightness (:obj:`int`): **[-100,100], default = 0**. |br|
            Adjusts the brightness of the image.
        contrast (:obj:`float`): **[1, 3], default = 1**. |br|
            Adjusts the contrast of the image.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Draws pose details onto input image.

        Args:
            inputs (dict): Dictionary with keys "img".
        """
        img = cv2.convertScaleAbs(
            inputs["img"], alpha=self.contrast, beta=self.brightness
        )
        return {"img": img}



# Copyright 2021 AI Singapore

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Records the nodes outputs to a CSV file.
"""

import logging
import textwrap
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List

from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.output.utils.csvlogger import CSVLogger


class Node(AbstractNode):
    """Tracks user-specified parameters and outputs the results in a CSV file.

    Inputs:
        ``all`` (:obj:`List`): A placeholder that represents a flexible input.
        Actual inputs to be written into the CSV file can be configured in
        ``stats_to_track``.

    Outputs:
        |none|

    Configs:
        stats_to_track (:obj:`List`):
            **default = ["keypoints", "bboxes", "bbox_labels"]**. |br|
            Parameters to log into the CSV file. The chosen parameters must be
            present in the data pool.
        file_path (:obj:`str`):
            **default = "PeekingDuck/data/stats.csv"**. |br|
            Directory where CSV file is saved.
        logging_interval (:obj:`int`): **default = 1**. |br|
            Interval between each log, in terms of seconds.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

        self.logger = logging.getLogger(__name__)
        self.logging_interval = int(self.logging_interval)  # type: ignore
        self.file_path = Path(self.file_path)  # type: ignore
        # check if file_path has a '.csv' extension
        if self.file_path.suffix != ".csv":
            raise ValueError("Filepath must have a '.csv' extension.")

        self._file_path_datetime = self._append_datetime_file_path(self.file_path)
        self._stats_checked = False
        self.stats_to_track: List[str]
        self.csv_logger = CSVLogger(
            self._file_path_datetime, self.stats_to_track, self.logging_interval
        )

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Writes the current state of the tracked statistics into
        the csv file as a row entry

        Args:
            inputs (dict): The data pool of the pipeline.

        Returns:
            outputs: [None]
        """
        # reset and terminate when there are no more data
        if inputs["pipeline_end"]:
            self._reset()
            return {}

        if not self._stats_checked:
            self._check_tracked_stats(inputs)
            # self._stats_to_track might change after the check
            self.csv_logger = CSVLogger(
                self._file_path_datetime, self.stats_to_track, self.logging_interval
            )

        self.csv_logger.write(inputs, self.stats_to_track)

        return {}

    def _check_tracked_stats(self, inputs: Dict[str, Any]) -> None:
        """Checks whether user input statistics is present in the data pool
        of the pipeline. Statistics not present in data pool will be
        ignored and dropped.
        """
        valid = []
        invalid = []

        for stat in self.stats_to_track:
            if stat in inputs:
                valid.append(stat)
            else:
                invalid.append(stat)

        if not invalid:
            msg = textwrap.dedent(
                f"""\
                {invalid} are not valid outputs.
                Data pool only has this outputs: {list(inputs.keys())}
                Only {valid} will be logged in the csv file
                """
            )
            self.logger.warning(msg)

        # update stats_to_track with valid stats found in data pool
        self.stats_to_track = valid
        self._stats_checked = True

    def _reset(self) -> None:
        del self.csv_logger

        # initialize for use in run
        self._stats_checked = False

    @staticmethod
    def _append_datetime_file_path(file_path: Path) -> Path:
        """Append time stamp to the filename."""
        current_time = datetime.now()
        # output as '240621-15-09-13'
        time_str = current_time.strftime("%d%m%y-%H-%M-%S")

        # append timestamp to filename before extension
        # Format: filename_timestamp.extension
        file_path_with_timestamp = file_path.with_name(
            f"{file_path.stem}_{time_str}{file_path.suffix}"
        )

        return file_path_with_timestamp



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Fast Pose Estimation model.
"""

from typing import Any, Dict

from peekingduck.pipeline.nodes.model.posenetv1 import posenet_model
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Initialises a PoseNet model to detect human poses from an image.

    The PoseNet node is capable of detecting multiple human figures
    simultaneously per inference and for each detected human figure, 17
    keypoints are estimated. The keypoint indices table can be found
    :ref:`here <whole-body-keypoint-ids>`.

    Inputs:
        |img|

    Outputs:
        |bboxes|

        |keypoints|

        |keypoint_scores|

        |keypoint_conns|

        |bbox_labels|

    Configs:
        model_type (:obj:`str`):
            **{"resnet", "50", "75", "100"}, default="resnet"**. |br|
            Defines the backbone model for PoseNet.
        weights_parent_dir (:obj:`Optional[str]`): **default = null**. |br|
            Change the parent directory where weights will be stored by replacing
            ``null`` with an absolute path to the desired directory.
        resolution (:obj:`Dict`):
            **default = { height: 225, width: 225 }**. |br|
            Resolution of input array to PoseNet model.
        max_pose_detection (:obj:`int`): **default = 10**. |br|
            Maximum number of poses to be detected.
        score_threshold (:obj:`float`): **[0, 1], default = 0.4**. |br|
            Threshold to determine if detection should be returned

    References:
        PersonLab: Person Pose Estimation and Instance Segmentation with a
        Bottom-Up, Part-Based, Geometric Embedding Model:
        https://arxiv.org/abs/1803.08225

        Code adapted from https://github.com/rwightman/posenet-python
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.model = posenet_model.PoseNetModel(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """function that reads the image input and returns the bboxes
        of the specified objects chosen to be detected
        """
        bboxes, keypoints, keypoint_scores, keypoint_conns = self.model.predict(
            inputs["img"]
        )
        bbox_labels = ["Person"] * len(bboxes)
        outputs = {
            "bboxes": bboxes,
            "keypoints": keypoints,
            "keypoint_scores": keypoint_scores,
            "keypoint_conns": keypoint_conns,
            "bbox_labels": bbox_labels,
        }
        return outputs



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for the create node CLI command."""

import functools
import re
from pathlib import Path
from typing import Callable, Dict, Optional, Tuple, Union

import click


def create_config_and_script_files(created_paths: Dict[str, Path]) -> None:
    """Creates a config and script file at the provided paths. The contents of
    these files will be copied from the template config and script file.
    """
    created_paths["config"].parent.mkdir(parents=True, exist_ok=True)
    created_paths["script"].parent.mkdir(parents=True, exist_ok=True)
    template_paths = get_config_and_script_paths(
        Path(__file__).resolve().parents[1],
        "configs",
        ("pipeline", "nodes"),
        "node_template",
    )
    with open(template_paths["config"]) as template_file, open(
        created_paths["config"], "w"
    ) as outfile:
        outfile.write(template_file.read())
    with open(template_paths["script"]) as template_file, open(
        created_paths["script"], "w"
    ) as outfile:
        lines = template_file.readlines()
        start = -1
        for i, line in enumerate(lines):
            if line.startswith('"'):
                start = i
                break
        # In case we couldn't find a starting line
        start = max(0, start)
        outfile.writelines(lines[start:])


def ensure_relative_path(node_subdir: str) -> str:
    """Checks that the subdir path does not contain parent directory
    navigator (..), is not absolute, and is not "peekingduck/pipeline/nodes".
    """
    pkd_node_subdir = "peekingduck/pipeline/nodes"
    if ".." in node_subdir:
        raise click.exceptions.UsageError("Path cannot contain '..'!")
    if Path(node_subdir).is_absolute():
        raise click.exceptions.UsageError("Path cannot be absolute!")
    if node_subdir == pkd_node_subdir:
        raise click.exceptions.UsageError(f"Path cannot be '{pkd_node_subdir}'!")
    return node_subdir


def ensure_valid_name(node_dir: Path, node_type: str, node_name: str) -> str:
    """Checks the validity of the specified node_name. Also checks if it
    already exists.
    """
    if re.match(r"^[a-zA-Z][\w\-]*[^\W_]$", node_name) is None:
        raise click.exceptions.UsageError("Invalid node name!")
    if (node_dir / node_type / f"{node_name}.py").exists():
        raise click.exceptions.UsageError("Node name already exists!")
    return node_name


def ensure_valid_name_partial(node_dir: Path, node_type: str) -> Callable:
    """Partial function to ensure_valid_name to provide a function that matches
    function signature required by ``value_proc`` in ``click.prompt()``.
    """
    return functools.partial(ensure_valid_name, node_dir, node_type)


def get_config_and_script_paths(
    parent_dir: Path,
    config_subdir: Union[str, Tuple[str, ...]],
    script_subdir: Union[str, Tuple[str, ...]],
    file_stem: str,
) -> Dict[str, Path]:
    """Returns the node config file and its corresponding script file."""
    if isinstance(config_subdir, tuple):
        config_subpath = Path(*config_subdir)
    else:
        config_subpath = Path(config_subdir)
    if isinstance(script_subdir, tuple):
        script_subpath = Path(*script_subdir)
    else:
        script_subpath = Path(script_subdir)

    return {
        "config": parent_dir / config_subpath / f"{file_stem}.yml",
        "script": parent_dir / script_subpath / f"{file_stem}.py",
    }


def verify_option(value: Optional[str], value_proc: Callable) -> Optional[str]:
    """Verifies that input value via click.option matches the expected value.

    This sets ``value`` to ``None`` if it is invalid so the rest of the prompt
    can flow smoothly.

    Args:
        value (Optional[str]): Input value.
        value_proc (Callable): A function to check the validity of ``value``.

    Returns:
        (Optional[str]): ``value`` if it is a valid value. ``None`` if it is
            not.

    Raises:
        click.exceptions.UsageError: When ``value`` is invalid.
    """
    if value is None:
        return value
    try:
        value = value_proc(value)
    except click.exceptions.UsageError as error:
        click.echo(f"Error: {error.message}", err=True)
        value = None
    return value
